package net.sf.jack4j;

/*
Copyright (C) 2008 Ondrej Par

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

import java.lang.ref.WeakReference;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * Client of Jack daemon.
 * 
 * <p>
 * The Java callback methods defined in this class (such as
 * {@link #process(int)}, {@link #xRunCallback()} etc.) <b>will only work if
 * both default native JackThreadInit callback and the default native callback
 * for the event are set</b>. For example, the <code>process(int)</code>
 * method is called only if you call {@link #setDefaultThreadInitCallback()} and
 * {@link #setDefaultProcessCallback()} before activating the client.
 * 
 * <p>
 * See {@link #setAllDefaultCallbacks()} if you want to set all native callbacks
 * at once.
 * 
 * <p>
 * Most methods correspond to functions defined in Jack library, except for
 * {@link #localPort(String)} and friends, that work with the internal Java
 * structure.
 * 
 * @author repa
 * 
 */
public abstract class JackClient {

	static {
		JackBridge.initializeJackBridge();
	}

	private volatile boolean isOpen;
	private volatile long clientHandle = 0;
	private volatile long callbackStruct = 0;
	private volatile boolean isActive;
	private Map<String, JackLocalPort> localPorts;
	private Map<Long, WeakReference<JackPort>> ports;
	private Map<Long, WeakReference<JackInternalClient>> internalClients;
	private JackClientThread clientThread;
	private final Statistics statistics;

	/**
	 * Creates new, inactive Jack client.
	 * 
	 * @param clientName
	 *            desired client name
	 * @param useExactName
	 *            if true, server won't modify the client name and exception is
	 *            thrown if client with that name already exists
	 * @param canStartServer
	 *            if true, the server will be started unless it's already
	 *            running
	 * @param serverName
	 *            if not null, connects to specified server; otherwise, connects
	 *            to default server
	 * @throws JackException
	 *             if communication with the server fails
	 */
	public JackClient(String clientName, boolean useExactName, boolean canStartServer, String serverName)
	        throws JackException {
		open(clientName, useExactName, canStartServer, serverName);

		isOpen = true;

		isActive = false;

		localPorts = new HashMap<String, JackLocalPort>();
		ports = new HashMap<Long, WeakReference<JackPort>>();

		internalClients = new HashMap<Long, WeakReference<JackInternalClient>>();

		clientThread = null;

		statistics = new Statistics(this);
	}

	/**
	 * @see java.lang.Object#finalize()
	 */
	@Override
	protected void finalize() throws Throwable {
		if (isOpen) {
			close();
		}
		super.finalize();
	}

	private native void open(String clientName, boolean useExactName, boolean canStartServer, String serverName)
	        throws JackException;

	/**
	 * Returns the pointer to C <code>jack_client_t</code> structure,
	 * represented as <code>long</code>.
	 */
	public long getClientHandle() {
		return clientHandle;
	}

	/**
	 * Returns the pointer to C structure that can be used by native Jack
	 * callbacks.
	 * 
	 * <p>
	 * The pointer is represented as <code>long</code> value.
	 */
	public long getCallbackStruct() {
		return callbackStruct;
	}

	/**
	 * Sets the default "on shutdown" callback for this client.
	 */
	public native void setDefaultShutdownCallback() throws JackException;

	/**
	 * Sets custom "on shutdown" callback for this client.
	 * 
	 * <p>
	 * The <code>pointer</code> must be a native
	 * <code>void(*)(void *arg)</code> pointer, represented as
	 * <code>long</code>.
	 * 
	 * @param pointer
	 *            the pointer to native function
	 * @param arg
	 *            the value passed as <code>void* arg</code> parameter to the
	 *            callback
	 * @see #setDefaultShutdownCallback()
	 */
	public native void setShutdownCallback(long pointer, long arg) throws JackException;

	/**
	 * Sets the default native JackProcessCallback for this client.
	 */
	public native void setDefaultProcessCallback() throws JackException;

	/**
	 * Sets custom JackProcessCallback for this client.
	 * 
	 * <p>
	 * The <code>pointer</code> must be a native
	 * <code>JackProcessCallback</code> pointer, represented as
	 * <code>long</code>.
	 * 
	 * @param pointer
	 *            the pointer to native function
	 * @param arg
	 *            the value passed as <code>void* arg</code> parameter to the
	 *            callback
	 * @see #setDefaultProcessCallback()
	 */
	public native void setProcessCallback(long pointer, long arg) throws JackException;

	/**
	 * Sets the default native JackThreadInitCallback for this client.
	 * 
	 * <p>
	 * This callback must be set, otherwise the other callbacks will not work
	 * correctly!
	 */
	public native void setDefaultThreadInitCallback() throws JackException;

	/**
	 * Sets custom JackThreadInitCallback for this client.
	 * 
	 * <p>
	 * The <code>pointer</code> must be a native
	 * <code>JackThreadInitCallback</code> pointer, represented as
	 * <code>long</code>.
	 * 
	 * @param pointer
	 *            the pointer to native function
	 * @param arg
	 *            the value passed as <code>void* arg</code> parameter to the
	 *            callback
	 * @see #setDefaultThreadInitCallback()
	 */
	public native void setThreadInitCallback(long pointer, long arg) throws JackException;

	/**
	 * Sets the default native JackFreewheelCallback for this client.
	 */
	public native void setDefaultFreewheelCallback() throws JackException;

	/**
	 * Sets custom JackFreewheelCallback for this client.
	 * 
	 * <p>
	 * The <code>pointer</code> must be a native
	 * <code>JackFreewheelCallback</code> pointer, represented as
	 * <code>long</code>.
	 * 
	 * @param pointer
	 *            the pointer to native function
	 * @param arg
	 *            the value passed as <code>void* arg</code> parameter to the
	 *            callback
	 * @see #setDefaultFreewheelCallback()
	 */
	public native void setFreewheelCallback(long pointer, long arg) throws JackException;

	/**
	 * Sets the default native JackBufferSizeCallback for this client.
	 */
	public native void setDefaultBufferSizeCallback() throws JackException;

	/**
	 * Sets custom JackBufferSizeCallback for this client.
	 * 
	 * <p>
	 * The <code>pointer</code> must be a native
	 * <code>JackBufferSizeCallback</code> pointer, represented as
	 * <code>long</code>.
	 * 
	 * @param pointer
	 *            the pointer to native function
	 * @param arg
	 *            the value passed as <code>void* arg</code> parameter to the
	 *            callback
	 * @see #setDefaultBufferSizeCallback()
	 */
	public native void setBufferSizeCallback(long pointer, long arg) throws JackException;

	/**
	 * Sets the default native JackSampleRateCallback for this client.
	 */
	public native void setDefaultSampleRateCallback() throws JackException;

	/**
	 * Sets custom JackSampleRateCallback for this client.
	 * 
	 * <p>
	 * The <code>pointer</code> must be a native
	 * <code>JackSampleRateCallback</code> pointer, represented as
	 * <code>long</code>.
	 * 
	 * @param pointer
	 *            the pointer to native function
	 * @param arg
	 *            the value passed as <code>void* arg</code> parameter to the
	 *            callback
	 * @see #setDefaultSampleRateCallback()
	 */
	public native void setSampleRateCallback(long pointer, long arg) throws JackException;

	/**
	 * Sets the default native JackClientRegistrationCallback for this client.
	 */
	public native void setDefaultClientRegistrationCallback() throws JackException;

	/**
	 * Sets custom JackClientRegistrationCallback for this client.
	 * 
	 * <p>
	 * The <code>pointer</code> must be a native
	 * <code>JackClientRegistrationCallback</code> pointer, represented as
	 * <code>long</code>.
	 * 
	 * @param pointer
	 *            the pointer to native function
	 * @param arg
	 *            the value passed as <code>void* arg</code> parameter to the
	 *            callback
	 * @see #setDefaultClientRegistrationCallback()
	 */
	public native void setClientRegistrationCallback(long pointer, long arg) throws JackException;

	/**
	 * Sets the default native JackPortRegistrationCallback for this client.
	 */
	public native void setDefaultPortRegistrationCallback() throws JackException;

	/**
	 * Sets custom JackPortRegistrationCallback for this client.
	 * 
	 * <p>
	 * The <code>pointer</code> must be a native
	 * <code>JackPortRegistrationCallback</code> pointer, represented as
	 * <code>long</code>.
	 * 
	 * @param pointer
	 *            the pointer to native function
	 * @param arg
	 *            the value passed as <code>void* arg</code> parameter to the
	 *            callback
	 * @see #setDefaultPortRegistrationCallback()
	 */
	public native void setPortRegistrationCallback(long pointer, long arg) throws JackException;

	/**
	 * Sets the default native JackPortConnectCallback for this client.
	 */
	public native void setDefaultPortConnectCallback() throws JackException;

	/**
	 * Sets custom JackPortConnectCallback for this client.
	 * 
	 * <p>
	 * The <code>pointer</code> must be a native
	 * <code>JackPortConnectCallback</code> pointer, represented as
	 * <code>long</code>.
	 * 
	 * @param pointer
	 *            the pointer to native function
	 * @param arg
	 *            the value passed as <code>void* arg</code> parameter to the
	 *            callback
	 * @see #setDefaultPortConnectCallback()
	 */
	public native void setPortConnectCallback(long pointer, long arg) throws JackException;

	/**
	 * Sets the default native JackGraphOrderCallback for this client.
	 */
	public native void setDefaultGraphOrderCallback() throws JackException;

	/**
	 * Sets custom JackGraphOrderCallback for this client.
	 * 
	 * <p>
	 * The <code>pointer</code> must be a native
	 * <code>JackGraphOrderCallback</code> pointer, represented as
	 * <code>long</code>.
	 * 
	 * @param pointer
	 *            the pointer to native function
	 * @param arg
	 *            the value passed as <code>void* arg</code> parameter to the
	 *            callback
	 * @see #setDefaultGraphOrderCallback()
	 */
	public native void setGraphOrderCallback(long pointer, long arg) throws JackException;

	/**
	 * Sets the default native JackXRunCallback for this client.
	 */
	public native void setDefaultXRunCallback() throws JackException;

	/**
	 * Sets custom JackXRunCallback for this client.
	 * 
	 * <p>
	 * The <code>pointer</code> must be a native <code>JackXRunCallback</code>
	 * pointer, represented as <code>long</code>.
	 * 
	 * @param pointer
	 *            the pointer to native function
	 * @param arg
	 *            the value passed as <code>void* arg</code> parameter to the
	 *            callback
	 * @see #setDefaultXRunCallback()
	 */
	public native void setXRunCallback(long pointer, long arg) throws JackException;

	/**
	 * Sets all available default native callbacks.
	 * 
	 * <p>
	 * The Java callback methods (such as {@link #process(int)}) will only work
	 * if the default native callback is set for thread initialization and for
	 * the corresponding event. It's recommended to call this method to set all
	 * those native callbacks at once.
	 * 
	 * <p>
	 * This method currently doesn't set the JackPortConnectCallback, because
	 * using that callback causes buggy behavior of Jackd 0.109.2.
	 */
	public void setAllDefaultCallbacks() throws JackException {
		setDefaultThreadInitCallback();

		setDefaultBufferSizeCallback();
		setDefaultClientRegistrationCallback();
		setDefaultFreewheelCallback();
		setDefaultGraphOrderCallback();
		// setDefaultPortConnectCallback(); // FIXME: commented out due to bug in Jackd 0.109.2
		setDefaultPortRegistrationCallback();
		setDefaultProcessCallback();
		setDefaultSampleRateCallback();
		setDefaultShutdownCallback();
		setDefaultXRunCallback();
	}

	/**
	 * Returns true if the client is still open.
	 */
	public boolean isOpen() {
		return isOpen;
	}

	/**
	 * Returns real client name (which may be different than the one specified
	 * in constructor).
	 */
	public native String getClientName() throws JackException;

	/**
	 * Closes the client.
	 * 
	 * <p>
	 * The client is completely unusable after this call.
	 */
	public native void close() throws JackException;

	/**
	 * Returns true if the server runs in real-time mode.
	 */
	public native boolean isRealtime() throws JackException;

	/**
	 * Returns true if the client was activated
	 */
	public boolean isActive() {
		return isActive;
	}

	/**
	 * Activates the client.
	 */
	public native void activate() throws JackException;

	/**
	 * Deactivates the client.
	 */
	public native void deactivate() throws JackException;

	/**
	 * Returns true if the Jack client thread was shut down.
	 * 
	 * <p>
	 * This method only works if the correct "on shutdown" callback was set,
	 * preferably using {@link #setDefaultShutdownCallback()}.
	 */
	public native boolean isShutdown();

	/**
	 * Calls <code>jack_thread_wait</code>.
	 */
	public native int threadWait(int status) throws JackException;

	/**
	 * Starts or stops Jack "freewheel" mode.
	 */
	public native void setFreewheel(boolean onoff) throws JackException;

	/**
	 * Sets buffer size used by Jack server.
	 * 
	 * <p>
	 * This can cause audio gap.
	 * 
	 * @throws IllegalArgumentException
	 *             if the bufferSize is not a power of two
	 * @throws JackException
	 *             if jack_set_buffer_size fails
	 */
	public native void setBufferSize(int bufferSize) throws JackException;

	/**
	 * Returns buffer size.
	 * 
	 * <p>
	 * Should only be used <b>before</b> {@link #activate()}. If you want to
	 * monitor buffer size changes, see {@link #bufferSizeCallback(int)}.
	 */
	public native int getBufferSize() throws JackException;

	/**
	 * Returns current sample rate.
	 */
	public native int getSampleRate() throws JackException;

	private native long registerPort(String portName, int flags, String portType, int bufferSize) throws JackException;

	/**
	 * Registers new port.
	 * 
	 * <p>
	 * Tip: you can to use {@link EnumSet} as <code>flags</code> parameter.
	 * 
	 * @param portName
	 *            short name of the new port
	 * @param flags
	 *            set of port flags
	 * @param portType
	 *            port type, currently only built-ins
	 *            {@link JackConstants#JACK_DEFAULT_AUDIO_TYPE()} and
	 *            {@link JackConstants#JACK_DEFAULT_MIDI_TYPE()} are allowed
	 * @param bufferSize
	 *            must be non-zero for non-standard portType, otherwise is
	 *            ignored
	 * @throws JackException
	 *             if port registration fails
	 */
	public synchronized JackLocalPort addPort(String portName, Collection<JackPortFlag> flags, String portType,
	        int bufferSize) throws JackException {
		EnumSet<JackPortFlag> flagSet = EnumSet.copyOf(flags);

		long portHandle = registerPort(portName, JackPortFlag.encode(flagSet), portType, bufferSize);

		JackLocalPort port;
		if (JackConstants.JACK_DEFAULT_MIDI_TYPE().equals(portType)) {
			port = new JackLocalMidiPort(portHandle, this);
		} else {
			port = new JackLocalPort(portHandle, this);
		}

		localPorts.put(portName, port);
		ports.put(port.getPortHandle(), new WeakReference<JackPort>(port));

		return port;
	}

	/**
	 * Registers new port of standard audio type.
	 * 
	 * <p>
	 * Calls {@link #addPort(String, Collection, String, int)} with portType set
	 * to {@link JackConstants#JACK_DEFAULT_AUDIO_TYPE()} and bufferSize set to
	 * 0.
	 */
	public synchronized JackLocalPort addAudioPort(String portName, Collection<JackPortFlag> flags)
	        throws JackException {
		return addPort(portName, flags, JackConstants.JACK_DEFAULT_AUDIO_TYPE(), 0);
	}

	/**
	 * Registers new port of standard MIDI type.
	 * 
	 * <p>
	 * Calls {@link #addPort(String, Collection, String, int)} with portType set
	 * to {@link JackConstants#JACK_DEFAULT_MIDI_TYPE()} and bufferSize set to
	 * 0.
	 */
	public synchronized JackLocalMidiPort addMidiPort(String portName, Collection<JackPortFlag> flags)
	        throws JackException {
		return (JackLocalMidiPort) addPort(portName, flags, JackConstants.JACK_DEFAULT_MIDI_TYPE(), 0);
	}

	/**
	 * Returns unmodifiable set of (short) names of ports registered by this
	 * client.
	 * 
	 * <p>
	 * The returned set will reflect changes (adding/removing ports), but cannot
	 * be modified with Set methods.
	 */
	public synchronized Set<String> localPortNames() {
		return Collections.unmodifiableSet(localPorts.keySet());
	}

	/**
	 * Returns port registered under given (short) name, or null if such port
	 * doesn't exist.
	 * 
	 * <p>
	 * Note that this method ignores aliases given to the ports.
	 */
	public synchronized JackLocalPort localPort(String name) {
		return localPorts.get(name);
	}

	private native void unregisterPort(JackLocalPort port) throws JackException;

	/**
	 * Unregisters given port.
	 * 
	 * <p>
	 * The corresponding {@link JackPort} instance will be unusable after this
	 * call.
	 * 
	 * @throws IllegalArgumentException
	 *             if port with given name doesn't exist
	 * @throws JackException
	 *             if the port couldn't be unregistered
	 */
	public synchronized void removePort(String portName) throws JackException {
		JackLocalPort port = localPorts.get(portName);
		if (port == null) {
			throw new IllegalArgumentException("JackPort doesn't exist: " + portName); //$NON-NLS-1$
		}

		long portHandle = port.getPortHandle();

		unregisterPort(port);

		localPorts.remove(portName);
		ports.remove(portHandle);
	}

	private synchronized JackPort portForHandle(long portHandle) {
		if (portHandle == 0) {
			return null;
		}

		WeakReference<JackPort> portReference = ports.get(portHandle);
		JackPort port = (portReference == null) ? null : portReference.get();

		if (port == null) {
			port = new JackPort(portHandle);
			ports.put(portHandle, new WeakReference<JackPort>(port));
		}

		return port;
	}

	/**
	 * Returns (local or foreign) port with given name, or null if it doesn't
	 * exist.
	 * 
	 * @param name
	 *            in the form 'client_name:form_name'
	 */
	public JackPort portByName(String name) throws JackException {
		return portForHandle(portByNameInternal(name));
	}

	private native long portByNameInternal(String name) throws JackException;

	/**
	 * Returns (local or foreign) port with given ID, or null if it doesn't
	 * exist.
	 * 
	 * <p>
	 * It seems that the only place where this method is currently usable are
	 * some of the callbacks, namely
	 * {@link #portRegistrationCallback(long, boolean)} and
	 * {@link #portConnectCallback(long, long, boolean)}.
	 * 
	 * @param id
	 *            unique ID of the port
	 */
	public JackPort portById(long id) throws JackException {
		return portForHandle(portByIdInternal(id));
	}

	private native long portByIdInternal(long id) throws JackException;

	/**
	 * Returns names of ports that fulfill given criteria.
	 * 
	 * @param portNamePattern
	 *            regex used to select ports by name; this parameter is ignored
	 *            if it's null or empty string
	 * @param typeNamePattern
	 *            regex used to select ports by type; this parameter is ignored
	 *            if it's null or empty string
	 * @param flagSet
	 *            required port flags; this parameter is ignored if it's null
	 * @return array of port names
	 */
	public String[] getPorts(String portNamePattern, String typeNamePattern, EnumSet<JackPortFlag> flagSet) {
		long flags;
		if (flagSet == null) {
			flags = 0;
		} else {
			flags = JackPortFlag.encode(flagSet);
		}

		return getPorts(portNamePattern, typeNamePattern, flags);
	}

	/**
	 * Returns names of ports that fulfill given criteria.
	 * 
	 * @param portNamePattern
	 *            regex used to select ports by name; this parameter is ignored
	 *            if it's null or empty string
	 * @param typeNamePattern
	 *            regex used to select ports by type; this parameter is ignored
	 *            if it's null or empty string
	 * @param flags
	 *            required port flags; this parameter is ignored if it's zero
	 * @return array of port names
	 */
	public native String[] getPorts(String portNamePattern, String typeNamePattern, long flags);

	/**
	 * Establish a connection between two ports.
	 * 
	 * <p>
	 * The port types must be identical, the source port must be an output port,
	 * and the destination port must be an input port.
	 * 
	 * <p>
	 * The ports need not to belong to this client.
	 * 
	 * @param sourcePortName
	 *            long name of source port
	 * @param destinationPortName
	 *            long name of destination port
	 * @return true if the connection was added, false if it already existed
	 * @throws JackException
	 *             if the connection can't be established
	 */
	public native boolean connect(String sourcePortName, String destinationPortName) throws JackException;

	/**
	 * Remove a connection between two ports.
	 * 
	 * <p>
	 * The port types must be identical, the source port must be an output port,
	 * and the destination port must be an input port.
	 * 
	 * <p>
	 * The ports need not to belong to this client.
	 * 
	 * @param sourcePortName
	 *            long name of source port
	 * @param destinationPortName
	 *            long name of destination port
	 * @throws JackException
	 *             if the connection can't be established
	 */
	public native void disconnect(String sourcePortName, String destinationPortName) throws JackException;

	/**
	 * Returns full port names of to which the specified port is connected.
	 * 
	 * This differs from {@link JackLocalPort#getConnections()} in two important
	 * respects:
	 * <ul>
	 * <li>You may not call this function from code that is executed in
	 * response to a JACK event. For example, you cannot use it in a
	 * GraphReordered handler.</li>
	 * <li>You need not be the owner of the port to get information about its
	 * connections.</li>
	 * </ul>
	 */
	public native String[] getAllPortConnections(JackPort port) throws JackException;

	/**
	 * Returns true if the specified port is owned by this client.
	 */
	public native boolean isMine(JackPort port) throws JackException;

	/**
	 * Recomputes port latencies.
	 */
	public native void recomputeTotalLatencies() throws JackException;

	/**
	 * Recomputes port latency.
	 */
	public native void recomputeTotalPortLatency(JackPort port) throws JackException;

	/**
	 * Returns port total latency.
	 */
	public native int getTotalPortLatency(JackPort port) throws JackException;

	/**
	 * If the port can be monitored, sets monitoring on or off.
	 */
	public native void requestPortMonitorByName(String portName, boolean on) throws JackException;

	/**
	 * Returns time in frames since Jack server began the current process cycle.
	 */
	public native int framesSinceCycleStart();

	/**
	 * Returns current frame time (running counter).
	 */
	public native int frameTime();

	/**
	 * Returns frame time after the last processing of the graph.
	 * 
	 * <p>
	 * This method should only be called from {@link #process(int)} callback.
	 */
	public native int lastFrameTime();

	/**
	 * Converts time in microseconds to frame time.
	 */
	public native int timeToFrames(long microseconds);

	/**
	 * Converts time in microseconds to frame time.
	 */
	public native long framesToTime(int frameTime);

	/**
	 * Returns current time in microseconds.
	 */
	public native long time();

	/**
	 * Returns current CPU load.
	 */
	public native float cpuLoad();

	/**
	 * Returns the {@link Statistics} object associated with this client.
	 */
	public Statistics getStatistics() {
		return statistics;
	}

	/**
	 * Returns the object that allows operations on client thread.
	 * 
	 * <p>
	 * Returns null if the thread wasn't created yet.
	 * 
	 * <p>
	 * The {@link #setDefaultThreadInitCallback()} must be used, otherwise this
	 * method always returns null.
	 */
	public JackClientThread getClientThread() {
		return clientThread;
	}

	/**
	 * Loads internal client.
	 * 
	 * @param clientName
	 *            name of internal client
	 * @param useExactName
	 *            if set to true, request that the client have exact clientName;
	 *            otherwise, Jack server may generate unique name
	 * @param sharedObjectName
	 *            if not null, specifies name of shared library; otherwise,
	 *            clientName is used
	 * @param initValue
	 *            if not null, this string will be passed to internal client's
	 *            <code>jack_initialize</code> function
	 */
	public JackInternalClient loadInternalClient(String clientName, boolean useExactName, String sharedObjectName,
	        String initValue) throws JackException {
		return internalClientForHandle(loadInternalClientNative(clientName, useExactName, sharedObjectName, initValue));
	}

	private native long loadInternalClientNative(String clientName, boolean useExactName, String sharedObjectName,
	        String initValue) throws JackException;

	/**
	 * Returns an object that represents internal client with specified name.
	 * 
	 * @throws JackException
	 *             if the client doesn't exist or if other error occurred
	 */
	public JackInternalClient internalClientByName(String clientName) throws JackException {
		return internalClientForHandle(internalClientByNameNative(clientName));
	}

	private native long internalClientByNameNative(String clientName) throws JackException;

	private synchronized JackInternalClient internalClientForHandle(long handle) {
		if (handle == 0) {
			return null;
		}

		WeakReference<JackInternalClient> internalClientReference = internalClients.get(handle);

		JackInternalClient internalClient = (internalClientReference == null) ? null : internalClientReference.get();

		if (internalClient == null) {
			internalClient = new JackInternalClient(clientHandle, handle);
			internalClients.put(handle, new WeakReference<JackInternalClient>(internalClient));
		}

		return internalClient;
	}

	/**
	 * Called during each process cycle.
	 * 
	 * <p>
	 * Must return zero on success, non-zero on error. If this method throws an
	 * exception, status -1 is returned to the server.
	 */
	public abstract int process(int bufferSize) throws Exception;

	/**
	 * Called once, when Jack server initializes client thread.
	 * 
	 * <p>
	 * Exceptions thrown by this callback are ignored.
	 */
	public abstract void threadInitCallback() throws Exception;

	/**
	 * Called when Jack enters or leaves "freewheel" mode.
	 * 
	 * <p>
	 * Exceptions thrown by this callback are ignored.
	 */
	public abstract void freewheelCallback(boolean onoff) throws Exception;

	/**
	 * Called when Jack buffer size changes.
	 * 
	 * <p>
	 * Must return zero on success, non-zero on error. If this method throws an
	 * exception, status -1 is returned to the server.
	 */
	public abstract int bufferSizeCallback(int newBufferSize) throws Exception;

	/**
	 * Called when Jack sample rate changes.
	 * 
	 * <p>
	 * Must return zero on success, non-zero on error. If this method throws an
	 * exception, status -1 is returned to the server.
	 */
	public abstract int sampleRateCallback(int newSampleRate) throws Exception;

	/**
	 * Called when another Jack client is registered/unregistered.
	 * 
	 * <p>
	 * Exceptions thrown by this callback are ignored.
	 * 
	 * @param clientName
	 *            name of the other client
	 * @param registered
	 *            true if the client is registered, false if it's unregistered
	 */
	public abstract void clientRegistrationCallback(String clientName, boolean registered) throws Exception;

	/**
	 * Called when Jack port is registered/unregistered.
	 * 
	 * <p>
	 * Exceptions thrown by this callback are ignored.
	 * 
	 * @param portId
	 *            unique port ID (see {@link #portById(long)})
	 * @param registered
	 *            true if the port is registered, false if it's unregistered
	 */
	public abstract void portRegistrationCallback(long portId, boolean registered) throws Exception;

	/**
	 * CURRENTLY NOT WORKING.
	 * 
	 * Called when two ports are connected or disconnected.
	 * 
	 * <p>
	 * Exceptions thrown by this callback are ignored.
	 * 
	 * <p>
	 * The Jack API documentation doesn't specify which port is source and which
	 * is destination.
	 * 
	 * @param portAId
	 *            unique ID of one of two ports connected or disconnected
	 * @param portBId
	 *            unique ID of one of two ports connected or disconnected
	 * @param connected
	 *            true if the connection is established, false if it's removed
	 * 
	 * @see #portById(long)
	 */
	public abstract void portConnectCallback(long portAId, long portBId, boolean connected) throws Exception;

	/**
	 * Called whenever the processing graph is reordered.
	 * 
	 * <p>
	 * Must return zero on success, non-zero on error. If this method throws an
	 * exception, status -1 is returned to the server.
	 */
	public abstract int graphOrderCallback() throws Exception;

	/**
	 * Called when Jack XRUN occurs.
	 * 
	 * <p>
	 * Must return zero on success, non-zero on error. If this method throws an
	 * exception, status -1 is returned to the server.
	 */
	public abstract int xRunCallback() throws Exception;

	void localPortNameChanged(JackLocalPort jackLocalPort, String portName) {
		String oldName = null;
		for (Map.Entry<String, JackLocalPort> entry : localPorts.entrySet()) {
			if (entry.getValue() == jackLocalPort) {
				oldName = entry.getKey();
				break;
			}
		}

		localPorts.remove(oldName);
		localPorts.put(portName, jackLocalPort);
	}
}
