package net.sf.jack4j;

/*
Copyright (C) 2008 Ondrej Par

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.channels.Channel;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.WritableByteChannel;

/**
 * Thread-safe, lock-free ringbuffer.
 * 
 * <p>
 * The ringbuffer can be accessed by two threads simultaneously. One thread can
 * read from the buffer, while the other can write to it. The access to the
 * ringbuffer need not to be synchronized, However, there can be only one reader
 * and one writer thread, and their roles can't be interchanged.
 * 
 * <p>
 * <b>Note on synchronization:</b> the various read/write methods in this class
 * are not synchronized (lack of synchronization is the point of lock-free
 * buffers), but it's developer's responsibility to ensure that there's always
 * only one thread reading, and only one thread writing.
 * 
 * @author repa
 * 
 */
public class RingBuffer {

	static {
		JackBridge.initializeJackBridge();
	}

	private class RingBufferChannel implements Channel {

		protected RingBuffer ringBuffer;
		protected boolean closeFlag;

		private RingBufferChannel(RingBuffer ringBuffer) {
			this.ringBuffer = ringBuffer;
			this.closeFlag = false;
		}

		public void close() throws IOException {
			closeFlag = true;
		}

		public boolean isOpen() {
			return !closeFlag;
		}
	}

	private final class RingBufferReadableChannel extends RingBufferChannel implements ReadableByteChannel {

		private RingBufferReadableChannel(RingBuffer ringBuffer) {
			super(ringBuffer);
		}

		public int read(ByteBuffer dst) throws IOException {
			if (closeFlag) {
				throw new ClosedChannelException();
			}

			return ringBuffer.readBytesToBuffer(dst, dst.remaining());
		}

	}

	private final class RingBufferWritableChannel extends RingBufferChannel implements WritableByteChannel {

		private RingBufferWritableChannel(RingBuffer ringBuffer) {
			super(ringBuffer);
		}

		public int write(ByteBuffer src) throws IOException {
			if (closeFlag) {
				throw new ClosedChannelException();
			}

			return ringBuffer.writeBytesFromBuffer(src, src.remaining());
		}

	}

	private long ringbufferHandle;
	private boolean freeWhenFinalized;
	private final int capacity;

	private byte[] tempReadArray = null;
	private byte[] tempWriteArray = null;

	/**
	 * Creates new ringbuffer that can hold specified number of bytes.
	 * 
	 * <p>
	 * The size is in <i>bytes</i>, not samples!
	 * 
	 * @param size
	 *            size in bytes
	 * @param freeWhenFinalized
	 *            specifies whether {@link #free()} should be called during
	 *            finalization
	 */
	public RingBuffer(int size, boolean freeWhenFinalized) throws JackException {
		this.ringbufferHandle = allocateRingBuffer(size);
		this.capacity = size;
		this.freeWhenFinalized = freeWhenFinalized;
	}

	/**
	 * Creates new ringbuffer that can hold specified number of bytes.
	 * 
	 * <p>
	 * Equivalent to <code>RingBuffer(size, true)</code>
	 */
	public RingBuffer(int size) throws JackException {
		this(size, true);
	}

	/**
	 * Creates new Java object, using pre-allocated ringbuffer.
	 * 
	 * <p>
	 * This method is useful when you obtain pointer to
	 * <code>jack_ringbuffer_t</code> from another JNI call and want to build
	 * the Java object around that native ringbuffer.
	 * 
	 * @param ringbufferHandle
	 *            pointer to <code>jack_ringbuffer_t</code>, packed as
	 *            <code>long</code>
	 * @param capacity
	 *            total capacity of the ringbuffer
	 * @param freeWhenFinalized
	 *            specifies whether {@link #free()} should be called during
	 *            finalization
	 */
	public RingBuffer(long ringbufferHandle, int capacity, boolean freeWhenFinalized) {
		this.ringbufferHandle = ringbufferHandle;
		this.capacity = capacity;
		this.freeWhenFinalized = freeWhenFinalized;
	}

	/**
	 * Frees the ringbuffer, if it didn't happen before.
	 * 
	 * @see java.lang.Object#finalize()
	 */
	@Override
	protected void finalize() throws Throwable {
		if (freeWhenFinalized) {
			free();
		}

		super.finalize();
	}

	/**
	 * Frees the ringbuffer.
	 * 
	 * <p>
	 * This object will be unusable after the call.
	 * 
	 * <p>
	 * This method may be called multiple calls; second and subsequent calls
	 * have no effect.
	 */
	public void free() throws JackException {
		deallocateRingBuffer(ringbufferHandle);
		ringbufferHandle = 0;
	}

	/**
	 * Returns the native pointer to <code>jack_ringbuffer_t</code> structure.
	 * 
	 * <p>
	 * The pointer is packed as <code>long</code> value, and can be used in
	 * JNI calls.
	 */
	public long getRingbufferHandle() {
		return ringbufferHandle;
	}

	/**
	 * Returns the total capacity of the ringbuffer (in bytes).
	 */
	public int getCapacity() {
		return capacity;
	}

	/**
	 * Creates and returns {@link ReadableByteChannel} associated with this
	 * RingBuffer.
	 * 
	 * <p>
	 * It's recommended (faster) to use only direct buffers with the returned
	 * channel.
	 * 
	 * <p>
	 * The channel behaves as a non-blocking channel, that is, read method
	 * doesn't block and reads only as many bytes as is available for reading in
	 * the associated RingBuffer.
	 * 
	 * <p>
	 * <b>The channel must not be used by multiple threads simultaneously.</b>
	 * This also means that the <code>read</code> method of the returned
	 * channel doesn't fully conform to ReadableByteChannel specification - if
	 * another thread starts reading from the channel during read operation, it
	 * won't block and the results are unpredictable.
	 * 
	 * <p>
	 * Closing the channel doesn't affect underlying buffer. If the underlying
	 * buffer is freed, the channel's <code>read</code> method will throw
	 * exception.
	 */
	public ReadableByteChannel createReadableByteChannel() {
		return new RingBufferReadableChannel(this);
	}

	/**
	 * Creates and returns {@link WritableByteChannel} associated with this
	 * RingBuffer.
	 * 
	 * <p>
	 * It's recommended (faster) to use only direct buffers with the returned
	 * channel.
	 * 
	 * <p>
	 * The channel behaves as a non-blocking channel, that is, write method
	 * doesn't block and writes only as many bytes as is available for writing
	 * in the associated RingBuffer.
	 * 
	 * <p>
	 * <b>The channel must not be used by multiple threads simultaneously.</b>
	 * This also means that the <code>write</code> method of the returned
	 * channel doesn't fully conform to WritableByteChannel specification - if
	 * another thread starts writing to the channel during write operation, it
	 * won't block and the results are unpredictable.
	 * 
	 * <p>
	 * Closing the channel doesn't affect underlying buffer. If the underlying
	 * buffer is freed, the channel's <code>write</code> method will throw
	 * exception.
	 */
	public WritableByteChannel createWritableByteChannel() {
		return new RingBufferWritableChannel(this);
	}

	/**
	 * Returns the number of bytes available for reading.
	 */
	public native int getReadSpaceBytes();

	/**
	 * Returns the number of samples available for reading.
	 */
	public int getReadSpaceSamples() {
		return getReadSpaceBytes() / (Float.SIZE / 8);
	}

	/**
	 * Returns the number of bytes available for writing.
	 */
	public native int getWriteSpaceBytes();

	/**
	 * Returns the number of samples available for writing.
	 */
	public int getWriteSpaceSamples() {
		return getWriteSpaceBytes() / (Float.SIZE / 8);
	}

	/**
	 * Reads data from the ringbuffer and stores them into specified direct
	 * buffer, then optionally advances read pointer.
	 * 
	 * <p>
	 * This method behaves exactly as
	 * {@link #readBytesToBuffer(ByteBuffer, int)}, but the specified
	 * <code>buffer</code> must be a direct buffer, otherwise an exception is
	 * thrown.
	 * 
	 * @return number of bytes actually copied
	 */
	public native int readBytesToDirectBuffer(ByteBuffer buffer, int size, boolean advance);

	/**
	 * Reads data from the ringbuffer and stores them into specified direct
	 * buffer, then advances pointer.
	 * 
	 * <p>
	 * This is an equivalent to
	 * <code>readBytesToDirectBuffer(buffer, size, true)</code>
	 */
	public int readBytesToDirectBuffer(ByteBuffer buffer, int size) {
		return readBytesToDirectBuffer(buffer, size, true);
	}

	/**
	 * Reads data from the ringbuffer and stores them into specified direct
	 * buffer, then optionally advances read pointer.
	 * 
	 * <p>
	 * If the <code>buffer</code> is not direct, the method throws
	 * IllegalArgumentException.
	 * 
	 * <p>
	 * The number of samples read is a minimum of: <code>size</code>
	 * parameter, return value of <code>buffer.remaining()</code>, and return
	 * value of {@link #getReadSpaceSamples()}.
	 * 
	 * @return number of samples actually copied
	 */
	public native int readSamplesToDirectBuffer(FloatBuffer buffer, int size, boolean advance);

	/**
	 * Reads data from the ringbuffer and stores them into specified direct
	 * buffer, then advances pointer.
	 * 
	 * <p>
	 * This is an equivalent to
	 * <code>readSamplesToDirectBuffer(buffer, size, true)</code>
	 */
	public int readSamplesToDirectBuffer(FloatBuffer buffer, int size) {
		return readSamplesToDirectBuffer(buffer, size, true);
	}

	/**
	 * Reads data from the ringbuffer and stores them into specified direct
	 * buffer, then optionally advances read pointer.
	 * 
	 * <p>
	 * The number of bytes read is a minimum of: specified <code>size</code>
	 * parameter, available space in target <code>buffer</code>, and size of
	 * available data in the ringbuffer, as returned by
	 * {@link #getReadSpaceBytes()}.
	 * 
	 * @return number of bytes actually copied
	 */
	public int readBytesToBuffer(ByteBuffer buffer, int size, boolean advance) {
		if (buffer.isDirect()) {
			return readBytesToDirectBuffer(buffer, size, advance);
		} else {
			// assure that data will fit into target buffer and temporary array
			int requestSize = Math.min(buffer.remaining(), Math.min(size, capacity));

			byte[] array;
			int arrayOffset;

			if (buffer.hasArray()) {
				array = buffer.array();
				arrayOffset = buffer.arrayOffset() + buffer.position();
			} else {
				// we will use tempReadArray here;; it's responsibility of the caller to assure
				// that this method is only called by a single thread, thus it's ok to use 
				// instance variable as temporary variable
				if (tempReadArray == null) {
					tempReadArray = new byte[capacity];
				}
				array = tempReadArray;
				arrayOffset = 0;
			}

			int result = readBytesToArray(array, arrayOffset, requestSize, advance);

			if (buffer.hasArray()) {
				// data were written into buffer, update position
				buffer.position(buffer.position() + result);
			} else {
				// data were written into temporary array, copy into the buffer
				buffer.put(array, arrayOffset, result);
			}

			return result;
		}
	}

	/**
	 * Reads data from the ringbuffer and stores them into specified direct
	 * buffer, then optionally advances read pointer.
	 * 
	 * <p>
	 * This is equivalent to <code>readBytesToBuffer(buffer, size, true)</code>
	 */
	public int readBytesToBuffer(ByteBuffer buffer, int size) {
		return readBytesToBuffer(buffer, size, true);
	}

	/**
	 * Reads data from the ringbuffer and stores them into specified direct
	 * buffer, then optionally advances read pointer.
	 * 
	 * <p>
	 * The number of bytes read is a minimum of specified <code>size</code>
	 * parameter and size of available data in the ringbuffer, as returned by
	 * {@link #getReadSpaceBytes()}.
	 * 
	 * <p>
	 * <code>start</code> must be non-negative and not greater than
	 * <code>array.length</code>; <code>size</code> must be non-negative
	 * and not greater than <code>array.length - start</code>.
	 * 
	 * @param array
	 *            the array into which will the data be copied
	 * @param start
	 *            index of first element to overwrite
	 * @param size
	 *            number of requested bytes
	 * @param advance
	 *            if set to true, read pointer is advanced after read
	 * @return number of bytes copied
	 * @throws JackException
	 *             if internal error occurs
	 * @throws IndexOutOfBoundsException
	 *             if <code>start</code> or <code>size</code> are out of
	 *             bounds
	 */
	public native int readBytesToArray(byte[] array, int start, int size, boolean advance);

	/**
	 * Reads data from the ringbuffer and stores them into specified direct
	 * buffer, then advances read pointer.
	 * 
	 * <p>
	 * This is equivalent to
	 * <code>readBytesToArray(array, start, size, true)</code>
	 */
	public int readBytesToArray(byte[] array, int start, int size) {
		return readBytesToArray(array, start, size, true);
	}

	/**
	 * Advances read position in the ringbuffer, skipping some readable data.
	 * 
	 * @param size
	 *            number of bytes to skip
	 */
	public native void readAdvanceBytes(int size);

	/**
	 * Advances read position in the ringbuffer, skipping some readable data.
	 * 
	 * @param size
	 *            number of samples to skip
	 */
	public void readAdvanceSamples(int size) {
		readAdvanceBytes(size * (Float.SIZE / 8));
	}

	/**
	 * Creates and returns RingBufferVector that covers readable part of the
	 * buffer.
	 * 
	 * <p>
	 * See {@link RingBufferVector} description for details.
	 * 
	 * <p>
	 * The use of these vectors in Java is not very efficient.
	 * 
	 * @param byteOrder
	 *            requested byte order of the ByteBuffers in the returned vector
	 */
	public native RingBufferVector createReadVector(ByteOrder byteOrder);

	/**
	 * Writes from the specified direct buffer into ringbuffer.
	 * 
	 * <p>
	 * This method behaves exactly as
	 * {@link #writeBytesFromBuffer(ByteBuffer, int)}, but the specified
	 * <code>buffer</code> must be a direct buffer, otherwise an exception is
	 * thrown.
	 * 
	 * @return number of bytes actually copied
	 */
	public native int writeBytesFromDirectBuffer(ByteBuffer buffer, int size);

	/**
	 * Writes from the specified direct buffer into ringbuffer.
	 * 
	 * <p>
	 * If the <code>buffer</code> is not direct, the method throws
	 * IllegalArgumentException.
	 * 
	 * <p>
	 * The number of bytes written is a minimum of: <code>size</code>
	 * parameter, return value of <code>buffer.remaining()</code>, and return
	 * value of {@link #getWriteSpaceBytes()}.
	 * 
	 * @return number of samples actually copied
	 */
	public native int writeSamplesFromDirectBuffer(FloatBuffer buffer, int size);

	/**
	 * Writes from the specified direct buffer into ringbuffer.
	 * 
	 * <p>
	 * The number of bytes written is a minimum of: specified <code>size</code>
	 * parameter, number of available bytes in target <code>buffer</code>,
	 * and size of available data in the ringbuffer, as returned by
	 * {@link #getWriteSpaceBytes()}.
	 * 
	 * @return number of bytes actually copied
	 */
	public int writeBytesFromBuffer(ByteBuffer buffer, int size) {
		if (buffer.isDirect()) {
			return writeBytesFromDirectBuffer(buffer, size);
		} else {
			// assure there's enough space in the buffer and temporary array
			int requestSize = Math.min(buffer.remaining(), Math.min(size, capacity));

			int originalPosition = buffer.position();

			byte[] array;
			int arrayOffset;

			if (buffer.hasArray()) {
				array = buffer.array();
				arrayOffset = buffer.arrayOffset() + buffer.position();
			} else {
				// we will use tempWriteArray here;; it's responsibility of the caller to assure
				// that this method is only called by a single thread, thus it's ok to use 
				// instance variable as temporary variable
				if (tempWriteArray == null) {
					tempWriteArray = new byte[capacity];
				}
				array = tempWriteArray;
				arrayOffset = 0;

				buffer.get(array, arrayOffset, requestSize);
			}

			int result = writeBytesFromArray(array, arrayOffset, requestSize);

			buffer.position(originalPosition + result);

			return result;
		}
	}

	/**
	 * Writes from the specified array into ringbuffer.
	 * 
	 * <p>
	 * The number of bytes read is a minimum of specified <code>size</code>
	 * parameter and size of available data in the ringbuffer, as returned by
	 * {@link #getWriteSpaceBytes()}.
	 * 
	 * <p>
	 * <code>start</code> must be non-negative and not greater than
	 * <code>array.length</code>; <code>size</code> must be non-negative
	 * and not greater than <code>array.length - start</code>.
	 * 
	 * @param array
	 *            the array from which will the data be copied
	 * @param start
	 *            index of first element that holds the data
	 * @param size
	 *            requested number of bytes to copy
	 * @return number of bytes copied
	 * @throws IndexOutOfBoundsException
	 *             if <code>start</code> or <code>size</code> are out of
	 *             bounds
	 */
	public native int writeBytesFromArray(byte[] array, int start, int size);

	/**
	 * Advances write position in the ringbuffer, skipping some writable space.
	 * 
	 * @param size
	 *            number of bytes to skip
	 */
	public native void writeAdvanceBytes(int size);

	/**
	 * Advances write position in the ringbuffer, skipping some writable space.
	 * 
	 * @param size
	 *            number of samples to skip
	 */
	public void writeAdvanceSamples(int size) {
		writeAdvanceBytes(size * (Float.SIZE / 8));
	}

	/**
	 * Creates and returns RingBufferVector that covers writable part of the
	 * buffer.
	 * 
	 * <p>
	 * See {@link RingBufferVector} description for details.
	 * 
	 * <p>
	 * The use of these vectors in Java is not very efficient.
	 * 
	 * @param byteOrder
	 *            requested byte order of the ByteBuffers in the returned vector
	 */
	public native RingBufferVector createWriteVector(ByteOrder byteOrder);

	/**
	 * Resets (clears) the buffer.
	 * 
	 * <p>
	 * This method is not thread-safe.
	 */
	public native void reset();

	/**
	 * Attempts to lock the underlying native buffer in memory.
	 * 
	 * <p>
	 * You can call this if you want to use the ringbuffer during real-time
	 * callbacks and want to avoid swapping.
	 */
	public native void mlock() throws JackException;

	/**
	 * Returns true if the native ringbuffer will be freed when this Java object
	 * is finalized.
	 */
	public boolean getFreeWhenFinalized() {
		return freeWhenFinalized;
	}

	/**
	 * Sets the flag that says if the native ringbuffer will be freed when this
	 * Java object is finalized.
	 * 
	 * <p>
	 * Normally, the flag is set to true; setting it to false may be useful if
	 * you want to share the ringbuffer with native code that will survive the
	 * destruction of this Java object.
	 */
	public void setFreeWhenFinalized(boolean freeWhenFinalized) {
		this.freeWhenFinalized = freeWhenFinalized;
	}

	private static native long allocateRingBuffer(int size) throws JackException;

	private static native void deallocateRingBuffer(long handle) throws JackException;
}
