package net.sf.jack4j;

/*
Copyright (C) 2008 Ondrej Par

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

import java.util.EnumSet;

/**
 * Contains information about current state of Jack transport.
 * 
 * <p>
 * The structure contains fields listed below. Note that some of the fields may
 * be uninitialized; to obtain information about which fields have valid values,
 * see {@link #getValidBits()} or {@link #getValidBitsSet()} methods.
 * 
 * <dl>
 * <dt>usecs</dt>
 * <dd>time in microseconds, monotonic, free-rolling counter</dd>
 * <dt>frameRate</dt>
 * <dd>current frame rate (per second)</dd>
 * <dt>frame</dt>
 * <dd>frame number; this field is always present</dd>
 * <dt>validBits</dt>
 * <dd>indicates which fields contain valid values, decode using
 * {@link JackPositionBit}</dd>
 * <dt>validBitsSet</dt>
 * <dd>same as validBits, but decoded into more convenient EnumSet</dd>
 * <dt>bar</dt>
 * <dd>current bar</dd>
 * <dt>beat</dt>
 * <dd>current beat-within-bar</dd>
 * <dt>tick</dt>
 * <dd>current tick-within-beat</dd>
 * <dt>barStartTick</dt>
 * <dd></dd>
 * <dt>beatsPerBar</dt>
 * <dd>time signature "numerator"</dd>
 * <dt>beatType</dt>
 * <dd>time signature "denominator"</dd>
 * <dt>ticksPerBeat</dt>
 * <dd></dd>
 * <dt>beatsPerMinute</dt>
 * <dd></dd>
 * <dt>frameTime</dt>
 * <dd>current time in seconds</dd>
 * <dt>nextTime</dt>
 * <dd>next sequential frame_time (unless repositioned)</dd>
 * <dt>bbtOffset</dt>
 * <dd>bar-beat-tick offset (how many frames passed since the start of current
 * tick)</dd>
 * <dt>audioFramesPerVideoFrame</dt>
 * <dd>number of audio frames per video frame</dd>
 * <dt>videoOffset</dt>
 * <dd>audio frame at which the first video frame in this cycle occurs</dd>
 * </dl>
 * 
 * @author repa
 * 
 */
public class TransportPosition implements Cloneable {

	static {
		JackBridge.initializeJackBridge();
	}

	@SuppressWarnings("unused")
	private long positionHandle;
	private boolean deallocateWhenFinalized;

	TransportPosition(long positionHandle, boolean deallocateWhenFinalized) {
		this.positionHandle = positionHandle;
		this.deallocateWhenFinalized = deallocateWhenFinalized;
	}

	public TransportPosition() {
		this.positionHandle = allocateTransportPosition();
		this.deallocateWhenFinalized = true;
	}

	@Override
	protected void finalize() throws Throwable {
		if (deallocateWhenFinalized) {
			deallocateTransportPosition(positionHandle);
		}
		super.finalize();
	}

	/**
	 * @see java.lang.Object#clone()
	 */
	@Override
	protected TransportPosition clone() {
		TransportPosition that;
		try {
			that = (TransportPosition) super.clone();
		} catch (CloneNotSupportedException e) {
			// this never happens
			throw new RuntimeException(e);
		}

		that.positionHandle = allocateCopyOfTransportPosition(positionHandle);
		that.deallocateWhenFinalized = true;

		return that;
	}

	/**
	 * @return the usecs
	 */
	public native long getUsecs();

	/**
	 * @param usecs
	 *            the usecs to set
	 */
	public native void setUsecs(long usecs);

	/**
	 * @return the frameRate
	 */
	public native int getFrameRate();

	/**
	 * @param frameRate
	 *            the frameRate to set
	 */
	public native void setFrameRate(int frameRate);

	/**
	 * @return the frame
	 */
	public native int getFrame();

	/**
	 * @param frame
	 *            the frame to set
	 */
	public native void setFrame(int frame);

	/**
	 * @return the validBits
	 */
	public native int getValidBits();

	/**
	 * @param validBits
	 *            the validBits to set
	 */
	public native void setValidBits(int validBits);

	/**
	 * @return the validBitsSet
	 */
	public EnumSet<JackPositionBit> getValidBitsSet() {
		return JackPositionBit.decode(getValidBits());
	}

	/**
	 * @param validBitsSet
	 *            the validBitsSet to set
	 */
	public void setValidBitsSet(EnumSet<JackPositionBit> validBitsSet) {
		setValidBits(JackPositionBit.encode(validBitsSet));
	}

	/**
	 * @return the bar
	 */
	public native int getBar();

	/**
	 * @param bar
	 *            the bar to set
	 */
	public native void setBar(int bar);

	/**
	 * @return the beat
	 */
	public native int getBeat();

	/**
	 * @param beat
	 *            the beat to set
	 */
	public native void setBeat(int beat);

	/**
	 * @return the tick
	 */
	public native int getTick();

	/**
	 * @param tick
	 *            the tick to set
	 */
	public native void setTick(int tick);

	/**
	 * @return the barStartTick
	 */
	public native double getBarStartTick();

	/**
	 * @param barStartTick
	 *            the barStartTick to set
	 */
	public native void setBarStartTick(double barStartTick);

	/**
	 * @return the beatsPerBar
	 */
	public native float getBeatsPerBar();

	/**
	 * @param beatsPerBar
	 *            the beatsPerBar to set
	 */
	public native void setBeatsPerBar(float beatsPerBar);

	/**
	 * @return the beatType
	 */
	public native float getBeatType();

	/**
	 * @param beatType
	 *            the beatType to set
	 */
	public native void setBeatType(float beatType);

	/**
	 * @return the ticks_per_beat
	 */
	public native double getTicksPerBeat();

	/**
	 * @param ticks_per_beat
	 *            the ticks_per_beat to set
	 */
	public native void setTicksPerBeat(double ticks_per_beat);

	/**
	 * @return the beats_per_minute
	 */
	public native double getBeatsPerMinute();

	/**
	 * @param beats_per_minute
	 *            the beats_per_minute to set
	 */
	public native void setBeatsPerMinute(double beats_per_minute);

	/**
	 * @return the frameTime
	 */
	public native double getFrameTime();

	/**
	 * @param frameTime
	 *            the frameTime to set
	 */
	public native void setFrameTime(double frameTime);

	/**
	 * @return the nextTime
	 */
	public native double getNextTime();

	/**
	 * @param nextTime
	 *            the nextTime to set
	 */
	public native void setNextTime(double nextTime);

	/**
	 * @return the bbtOffset
	 */
	public native int getBbtOffset();

	/**
	 * @param bbtOffset
	 *            the bbtOffset to set
	 */
	public native void setBbtOffset(int bbtOffset);

	/**
	 * @return the audioFramesPerVideoFrame
	 */
	public native float getAudioFramesPerVideoFrame();

	/**
	 * @param audioFramesPerVideoFrame
	 *            the audioFramesPerVideoFrame to set
	 */
	public native void setAudioFramesPerVideoFrame(float audioFramesPerVideoFrame);

	/**
	 * @return the videoOffset
	 */
	public native int getVideoOffset();

	/**
	 * @param videoOffset
	 *            the videoOffset to set
	 */
	public native void setVideoOffset(int videoOffset);

	private static native long allocateTransportPosition();

	private static native long allocateCopyOfTransportPosition(long originalHandle);

	private static native void deallocateTransportPosition(long positionHandle);
}
