package net.sf.jack4j.util.file.wav;

/*
Copyright (C) 2008 Ondrej Par

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

import java.nio.ByteBuffer;

/**
 * Converts samples stored in WAVE file to <code>float</code> and vice-versa.
 * 
 * @author repa
 * 
 */
public abstract class WavSampleConverter {

	private final static class WavSampleConverter8BitUnsigned extends WavSampleConverter {

		@Override
		public int bytesPerSample() {
			return 1;
		}

		@Override
		public void floatToWavSample(float sample, ByteBuffer byteBuffer) {
			int value = (int) (sample * 128 + 128);
			if (value < 0) {
				value = 0;
			} else if (value > 255) {
				value = 255;
			}

			byteBuffer.put((byte) value);
		}

		@Override
		public float wavSampleToFloat(ByteBuffer byteBuffer) {
			return (byteBuffer.get() - 128) / 128.0f;
		}

	}

	private final static class WavSampleConverter16BitSigned extends WavSampleConverter {

		@Override
		public int bytesPerSample() {
			return 2;
		}

		@Override
		public void floatToWavSample(float sample, ByteBuffer byteBuffer) {
			int value = (int) (sample * 0x8000);
			byteBuffer.put((byte) (value & 0xff));
			byteBuffer.put((byte) ((value >> 8) & 0xff));
		}

		@Override
		public float wavSampleToFloat(ByteBuffer byteBuffer) {
			int intValue = ((byteBuffer.get() & 0xff) | (byteBuffer.get() << 8));
			return ((float) intValue) / 0x8000;
		}

	}

	private final static class WavSampleConverter24BitSigned extends WavSampleConverter {

		@Override
		public int bytesPerSample() {
			return 3;
		}

		@Override
		public void floatToWavSample(float sample, ByteBuffer byteBuffer) {
			int value = (int) (sample * 0x800000);
			byteBuffer.put((byte) (value & 0xff));
			byteBuffer.put((byte) ((value >> 8) & 0xff));
			byteBuffer.put((byte) ((value >> 16) & 0xff));
		}

		@Override
		public float wavSampleToFloat(ByteBuffer byteBuffer) {
			int intValue = ((byteBuffer.get() & 0xff) | (byteBuffer.get() << 8) | (byteBuffer.get() << 16));
			return ((float) intValue) / 0x800000;
		}

	}

	private final static class WavSampleConverter32BitSigned extends WavSampleConverter {

		@Override
		public int bytesPerSample() {
			return 4;
		}

		@Override
		public void floatToWavSample(float sample, ByteBuffer byteBuffer) {
			int value = (int) (sample * 0x80000000);
			byteBuffer.put((byte) (value & 0xff));
			byteBuffer.put((byte) ((value >> 8) & 0xff));
			byteBuffer.put((byte) ((value >> 16) & 0xff));
			byteBuffer.put((byte) ((value >> 24) & 0xff));
		}

		@Override
		public float wavSampleToFloat(ByteBuffer byteBuffer) {
			int intValue = ((byteBuffer.get() & 0xff) | (byteBuffer.get() << 8) | (byteBuffer.get() << 16) | (byteBuffer
			        .get() << 24));
			return ((float) intValue) / 0x80000000;
		}

	}

	public static WavSampleConverter createWavSampleConverter(int bitsPerSample) {
		if (bitsPerSample <= 0) {
			throw new IllegalArgumentException();
		} else if (bitsPerSample <= 8) {
			return new WavSampleConverter8BitUnsigned();
		} else if (bitsPerSample <= 16) {
			return new WavSampleConverter16BitSigned();
		} else if (bitsPerSample <= 24) {
			return new WavSampleConverter24BitSigned();
		} else if (bitsPerSample <= 32) {
			return new WavSampleConverter32BitSigned();
		} else {
			throw new IllegalArgumentException();
		}
	}

	public abstract int bytesPerSample();

	public abstract void floatToWavSample(float sample, ByteBuffer byteBuffer);

	public abstract float wavSampleToFloat(ByteBuffer byteBuffer);
}
